---------------------------------------------------------------------------
-- @author Uli Schlachter
-- @copyright 2010 Uli Schlachter
-- @module gears
---------------------------------------------------------------------------


return
{
    color = require("gears.color");
    debug = require("gears.debug");
    object = require("gears.object");
    surface = require("gears.surface");
    wallpaper = require("gears.wallpaper");
    timer = require("gears.timer");
    cache = require("gears.cache");
    matrix = require("gears.matrix");
    shape = require("gears.shape");
    protected_call = require("gears.protected_call");
    geometry = require("gears.geometry");
    math = require("gears.math");
    table = require("gears.table");
    string = require("gears.string");
    sort = require("gears.sort");
    filesystem = require("gears.filesystem");
}

-- vim: filetype=lua:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
